Imports System
Imports System.IO
Imports System.Windows.Forms
Imports System.Drawing
Imports DeckConfirmation
Imports Emgu.CV
Imports Emgu.CV.UI
Imports Emgu.CV.Structure
Imports Emgu.CV.CvEnum
Imports Emgu.CV.Reflection

Public Class CVisionExecutor
    Private Const CORRECT_NOTCH_ORIENTATION As String = "Correct orientation:"
    Private Const WRONG_NOTCH_ORIENTATION As String = "Wrong orientation:"
    Private Const NOTCH_LOWERRIGHT As Integer = 0
    Private Const NOTCH_UPPERRIGHT As Integer = 1
    Private Const NOTCH_LOWERLEFT As Integer = 2
    Private Const NOTCH_UPPERLEFT As Integer = 3
    Private Const MIN_HEIGHT_OF_TALL_RESERVOIR As Integer = 25

    Private Enum EnumNotchVals
        lr = 0
        ur
        ll
        ul
    End Enum

    Public _maestro As CaliperLifeSciences.VSTA.MaestroMacroHost.Application
    'Private _maestro As Maestro
    Private _valueClass As ValueClass
    Private _lightingEval As LightingEvaluation = New LightingEvaluation()

    Private _rfDD As RecFourDeckDetails
    Private _rf As RecFour = New RecFour()
    Private _td As TipDetection = New TipDetection()
    Private _rr As RecResult = New RecResult()
    Private _ph As Plate384Height = New Plate384Height()
    Private _nd As Notch = New Notch()
    Private _sd As SunkenDeck = New SunkenDeck()
    Private _heightOf384 As Integer
    Private _heightOfSunkenDeck() As Integer = New Integer(4) {}
    Private _notchResult As String
    Private _visionErrorInTipDetection As Boolean
    Private _brightness As Integer
    Private _capture As Capture

    ' initialize camera capture, need to be called at the beginning of each application
    Public Sub init(ByVal simMode As Boolean, ByVal maestro As Maestro)
        _valueClass = New ValueClass
        _valueClass.DiagnosticInputFromFiles = False
        _valueClass.DiagnosticInputToFiles = False
        _valueClass.DiagnosticShowDisplays = False
        _valueClass.DiagnosticFileFolder = "C:\\Vision_Bitmaps\\"

        _lightingEval.Init(_valueClass)

        _rf.Init(_valueClass)
        _td.Init(_valueClass)
        _ph.Init(_valueClass)
        _nd.Init(_valueClass)
        _sd.init(_valueClass)
        _rr.errors = New Boolean(4) {}
        _visionErrorInTipDetection = False
        If simMode = False Then
            _capture = New Capture()
        End If
        _maestro = maestro
    End Sub

    Public WriteOnly Property Brightness() As Integer
        Set(ByVal Value As Integer)
            _brightness = Value
        End Set
    End Property

    'recognize consumable types on four deck locations 
    'pass in expected types and riser situation
    Public Sub RecF( _
                    ByRef ulExpected As CConsumable, _
                    ByRef llExpected As CConsumable, _
                    ByRef urExpected As CConsumable, _
                    ByRef lrExpected As CConsumable, _
                    ByVal locationType As String(), _
                    ByVal numConsumables As Integer(), _
                    ByVal hasLid As Boolean())

        Dim ulExpId As Double = 0
        Dim llExpId As Double = 0
        Dim urExpId As Double = 0
        Dim lrExpId As Double = 0
        Dim ulExpColor As String = String.Empty
        Dim llExpColor As String = String.Empty
        Dim urExpColor As String = String.Empty
        Dim lrExpColor As String = String.Empty

        ' transform expected consumable into an intermediate id
        EstablishExpConsIdAndColor(ulExpected, ulExpId, ulExpColor)
        EstablishExpConsIdAndColor(llExpected, llExpId, llExpColor)
        EstablishExpConsIdAndColor(urExpected, urExpId, urExpColor)
        EstablishExpConsIdAndColor(lrExpected, lrExpId, lrExpColor)

        _rf.SetExpectedData(lrExpId, urExpId, llExpId, ulExpId, lrExpColor, urExpColor, llExpColor, ulExpColor)

        _rfDD = New RecFourDeckDetails(locationType, numConsumables, hasLid)

        ' use the intermediate id to do recognition
        Dim res() As Double = _rf.Run(True, _brightness, _capture, _rfDD)

        Dim idx As Integer
        Dim idx180 As Integer
        For idx = 0 To 3
            'reduce the 96 plate stack height by 1 or 2
            If (res(idx) >= 1.1 And res(idx) <= 1.4) Then
                ' from the perspective of the camera (180 from front of machine)
                If idx = 0 Then
                    idx180 = 3
                ElseIf idx = 1 Then
                    idx180 = 2
                ElseIf idx = 2 Then
                    idx180 = 1
                ElseIf idx = 3 Then
                    idx180 = 0
                End If
                If _rfDD.LocationType(idx180) = StaticClass.LOCATIONTYPE_CPAC_MICROPLATE Then ' especially tall
                    res(idx) = res(idx) - 0.2
                ElseIf Not _rfDD.LocationType(idx180) = StaticClass.LOCATIONTYPE_PLATELOCATOR Then
                    res(idx) = res(idx) - 0.1
                End If
            End If
        Next

        'transform the consumable id to string
        Dim ulActual As String = Consumable(res(3))
        Dim llActual As String = Consumable(res(2))
        Dim urActual As String = Consumable(res(1))
        Dim lrActual As String = Consumable(res(0))
        'special deck locations where there is no consumable specified
        ' TODO: Verify that this code should be removed (what's there is there)
        If ulExpected Is Nothing Then
            ulActual = String.Empty
        End If
        If llExpected Is Nothing Then
            llActual = String.Empty
        End If
        If urExpected Is Nothing Then
            urActual = String.Empty
        End If
        If lrExpected Is Nothing Then
            lrActual = String.Empty
        End If
        'store the expected values and actual values in 'rr' to be used by diagnostic form
        _rr.ulActual = ulActual
        _rr.llActual = llActual
        _rr.urActual = urActual
        _rr.lrActual = lrActual

        _rr.ulExpected = String.Empty
        _rr.llExpected = String.Empty
        _rr.urExpected = String.Empty
        _rr.lrExpected = String.Empty
        If Not ulExpected Is Nothing Then
            _rr.ulExpected = ulExpected.Name
        End If
        If Not llExpected Is Nothing Then
            _rr.llExpected = llExpected.Name
        End If
        If Not urExpected Is Nothing Then
            _rr.urExpected = urExpected.Name
        End If
        If Not lrExpected Is Nothing Then
            _rr.lrExpected = lrExpected.Name
        End If

        _rr.errors(0) = False
        _rr.errors(1) = False
        _rr.errors(2) = False
        _rr.errors(3) = False
        ' compare actual vs expected consumable types at each of the 4 deck locations;
        ' indicate which (if any) of the four locations doesn't match;
        ' there's an error if they don't match and no riser is present
        If Not ExpectedAndActualMatch(ulExpected, ulActual, numConsumables(0)) Then 'And riserType(0) = String.Empty Then
            _rr.errors(0) = True
        End If
        If Not ExpectedAndActualMatch(llExpected, llActual, numConsumables(1)) Then ' And riserType(1) = String.Empty Then
            _rr.errors(1) = True
        End If
        If Not ExpectedAndActualMatch(urExpected, urActual, numConsumables(2)) Then ' And riserType(2) = String.Empty Then
            _rr.errors(2) = True
        End If
        If Not ExpectedAndActualMatch(lrExpected, lrActual, numConsumables(3)) Then ' And riserType(3) = String.Empty Then
            _rr.errors(3) = True
        End If

    End Sub

    Private Sub EstablishExpConsIdAndColor(ByRef expCons As CConsumable, ByRef expId As Double, ByRef expColor As String)
        expId = -1 'StaticClass.NO_CONSUMABLE
        expColor = String.Empty
        If Not expCons Is Nothing Then
            If expCons.Category = "Tip Box" Then
                expId = StaticClass.TIPBOX
            ElseIf expCons.Category = "SPE Plate" Or _
               expCons.Category = "Collection Plate" Or _
               expCons.Category = "SBS Plate" Then
                If expCons.Density = 384 Then
                    expId = StaticClass.PLATE_384
                ElseIf expCons.Density = 96 Then
                    expId = StaticClass.PLATE_96
                End If
            ElseIf expCons.Category = "Reservoir" Then
                If expCons.Density = 4 Then
                    expId = StaticClass.RESERVOIR_4
                ElseIf expCons.Density = 12 Then
                    expId = StaticClass.RESERVOIR_12
                ElseIf expCons.Density = 1 Then
                    expId = StaticClass.RESERVOIR_1
                End If
            ElseIf expCons.Category = "Deep Well" Then
                If expCons.WellShape = "Square" Then
                    expId = StaticClass.PLATE_96DEEPWELL_SQUARE
                Else
                    expId = StaticClass.PLATE_96DEEPWELL_CIRCLE
                End If
            End If

            expColor = expCons.Color
        End If
    End Sub

    Private Function ExpectedAndActualMatch(ByRef expCons As CConsumable, ByVal actualVal As String, ByVal numConsumable As Integer) As Boolean
        Dim retVal As Boolean = False
        Dim expVal As String = String.Empty

        If Not expCons Is Nothing Then
            If expCons.Category = "Tip Box" Then
                expVal = "Tip Box " & expCons.Density.ToString()
            ElseIf expCons.Category = "SPE Plate" Or _
               expCons.Category = "Collection Plate" Or _
               expCons.Category = "SBS Plate" Then
                expVal = "Plate " & expCons.Density.ToString()
            ElseIf expCons.Category = "Reservoir" Then
                expVal = "Reservoir " & expCons.Density.ToString()
            ElseIf expCons.Category = "Deep Well" Then
                expVal = "Deep Well Plate " & expCons.Density.ToString()
            End If
        End If

        If actualVal = "As Expected, No Consumable" Then
            retVal = True
        ElseIf expVal = "Plate 96" Then
            If numConsumable > 1 Then
                If actualVal = (numConsumable.ToString() & " * Plate 96") Then
                    retVal = True
                End If
            Else
                If actualVal = "Plate 96" Then
                    retVal = True
                End If
            End If
        Else
            If actualVal = expVal Then
                retVal = True
            End If
        End If

        Return retVal
    End Function

    'detect tips in a tip box
    'expectation is that every Tip Box will contain ALL 96 tips
    Public Function TipDetect() As Integer
        Dim result As Integer = _td.Run(True, _brightness, _capture)

        _visionErrorInTipDetection = False
        If Not result = 0 Then
            ' 0 == match; 2 == not a match; -1 == something else went wrong
            _visionErrorInTipDetection = True
        End If
    End Function

    'the methods below detect notch locations on different consumables
    'the correct notch positions are set up based on the consumables used during development, to be changed if plates from different vendors are used
    Public Sub NotchDetect_96(ByVal expectedNotch As String, ByVal numStackedConsumables As Integer)
        Dim actualPosition As Integer = _nd.Run(True, StaticClass.PLATE_96, ConvertNotchString(expectedNotch), _brightness, _capture, numStackedConsumables)
        Dim enumActualVal As EnumNotchVals = CType(actualPosition, EnumNotchVals)

        If (enumActualVal.ToString() = expectedNotch) Then
            NotchMsg(ConvertNotchString(expectedNotch), CORRECT_NOTCH_ORIENTATION)
        Else
            NotchMsg(ConvertNotchString(expectedNotch), WRONG_NOTCH_ORIENTATION)
        End If
    End Sub

    Public Sub NotchDetect_96DWP(ByVal expectedNotch As String, ByVal expCons As CConsumable)
        Dim actualPosition As Integer = -1

        If expCons.WellShape = "Square" Then
            actualPosition = _nd.Run(True, StaticClass.PLATE_96DEEPWELL_SQUARE, ConvertNotchString(expectedNotch), _brightness, _capture, 1)
        Else
            actualPosition = _nd.Run(True, StaticClass.PLATE_96DEEPWELL_CIRCLE, ConvertNotchString(expectedNotch), _brightness, _capture, 1)
        End If

        Dim enumActualVal As EnumNotchVals = CType(actualPosition, EnumNotchVals)
        If (enumActualVal.ToString() = expectedNotch) Then
            NotchMsg(ConvertNotchString(expectedNotch), CORRECT_NOTCH_ORIENTATION)
        Else
            NotchMsg(ConvertNotchString(expectedNotch), WRONG_NOTCH_ORIENTATION)
        End If
    End Sub

    Public Sub NotchDetect_384(ByVal expectedNotch As String, ByVal numStackedConsumables As Integer)
        Dim actualPosition As Integer = _nd.Run(True, StaticClass.PLATE_384, ConvertNotchString(expectedNotch), _brightness, _capture, numStackedConsumables)
        Dim enumActualVal As EnumNotchVals = CType(actualPosition, EnumNotchVals)

        If (enumActualVal.ToString() = expectedNotch) Then
            NotchMsg(ConvertNotchString(expectedNotch), CORRECT_NOTCH_ORIENTATION)
        Else
            NotchMsg(ConvertNotchString(expectedNotch), WRONG_NOTCH_ORIENTATION)
        End If
    End Sub

    Public Sub NotchDetect_4Well(ByVal expectedNotch As String, ByVal expCons As CConsumable)
        Dim actualPosition As Integer = -1

        If expCons.Height >= MIN_HEIGHT_OF_TALL_RESERVOIR Then
            actualPosition = _nd.Run(True, StaticClass.RESERVOIR_4TALL, ConvertNotchString(expectedNotch), _brightness, _capture, 1)
        Else
            actualPosition = _nd.Run(True, StaticClass.RESERVOIR_4, ConvertNotchString(expectedNotch), _brightness, _capture, 1)
        End If

        Dim enumActualVal As EnumNotchVals = CType(actualPosition, EnumNotchVals)
        If (enumActualVal.ToString() = expectedNotch) Then
            NotchMsg(ConvertNotchString(expectedNotch), CORRECT_NOTCH_ORIENTATION)
        Else
            NotchMsg(ConvertNotchString(expectedNotch), WRONG_NOTCH_ORIENTATION)
        End If
    End Sub

    Public Sub NotchDetect_12Well(ByVal expectedNotch As String, ByRef expCons As CConsumable)
        Dim actualPosition As Integer = -1

        If expCons.Height >= MIN_HEIGHT_OF_TALL_RESERVOIR Then
            actualPosition = _nd.Run(True, StaticClass.RESERVOIR_12TALL, ConvertNotchString(expectedNotch), _brightness, _capture, 1)
        Else
            actualPosition = _nd.Run(True, StaticClass.RESERVOIR_12, ConvertNotchString(expectedNotch), _brightness, _capture, 1)
        End If

        Dim enumActualVal As EnumNotchVals = CType(actualPosition, EnumNotchVals)
        If (enumActualVal.ToString() = expectedNotch) Then
            NotchMsg(ConvertNotchString(expectedNotch), CORRECT_NOTCH_ORIENTATION)
        Else
            NotchMsg(ConvertNotchString(expectedNotch), WRONG_NOTCH_ORIENTATION)
        End If
    End Sub

    Private Function ConvertNotchString(ByVal notchStr As String) As Integer
        Dim retVal As Integer = -1

        ' Note: This logic is performed from the orientation of the camera (i.e. from
        ' the back of the machine); the picture has not been rotated 180 degress yet
        If notchStr = "ul" Then
            retVal = 3
        ElseIf notchStr = "ll" Then
            retVal = 2
        ElseIf notchStr = "ur" Then
            retVal = 1
        ElseIf notchStr = "lr" Then
            retVal = 0
        End If
        Return retVal
    End Function

    'translate notch position id (0-3) to actual positions in strings and combine with the result (correct/incorrect orientation) passed in
    'store the combined string in notchResult to be display in the diagnostic form
    Public Sub NotchMsg(ByVal notchPos As Integer, ByVal result As String)
        Dim msg As String = result + vbCrLf

        If result = CORRECT_NOTCH_ORIENTATION Then
            If (notchPos = NOTCH_UPPERLEFT) Then
                msg += "Notch in the upper left corner."
            ElseIf (notchPos = NOTCH_LOWERLEFT) Then
                msg += "Notch in the lower left corner."
            ElseIf (notchPos = NOTCH_UPPERRIGHT) Then
                msg += "Notch in the upper right corner."
            ElseIf (notchPos = NOTCH_LOWERRIGHT) Then
                msg += "Notch in the lower right corner."
            End If
        Else ' wrong orientation
            If (notchPos = NOTCH_UPPERLEFT) Then
                msg += "Notch expected in the upper left corner."
            ElseIf (notchPos = NOTCH_LOWERLEFT) Then
                msg += "Notch expected in the lower left corner."
            ElseIf (notchPos = NOTCH_UPPERRIGHT) Then
                msg += "Notch expected in the upper right corner."
            ElseIf (notchPos = NOTCH_LOWERRIGHT) Then
                msg += "Notch expected in the lower right corner."
            End If
        End If
        _notchResult = msg
    End Sub

    'check the height of 384 plate stack
    Public Sub CalcHeight(ByVal deckLoc As String)
        Try
            _heightOf384 = _ph.Run(True, _brightness, _capture)

            Dim numStackedConsumables As Integer = _maestro.GetStackedConsumablesAtLocation(deckLoc)
            If Not _heightOf384 = numStackedConsumables Then
                _rr.errors(0) = True
            End If
        Catch ex As Exception
            MsgBox("Error in CalcHeight(): " & ex.Message)
        End Try
    End Sub

    'check the heights of tip box stack in the sunken deck
    Public Sub SunkenDeck(ByVal deckLoc As String, ByRef ul As CConsumable, ByRef ll As CConsumable, ByRef ur As CConsumable, ByRef lr As CConsumable)
        _heightOfSunkenDeck = _sd.Run(True, _brightness, _capture)

        Dim numStackedConsumables As Integer
        If deckLoc = "A1" Then
            numStackedConsumables = _maestro.GetStackedConsumablesAtLocation("A0")
            If Not _heightOfSunkenDeck(0) = numStackedConsumables Then
                _rr.errors(0) = True
            End If
            numStackedConsumables = _maestro.GetStackedConsumablesAtLocation("B0")
            If Not _heightOfSunkenDeck(1) = numStackedConsumables Then
                _rr.errors(1) = True
            End If
            numStackedConsumables = _maestro.GetStackedConsumablesAtLocation("A1")
            If Not _heightOfSunkenDeck(2) = numStackedConsumables Then
                _rr.errors(2) = True
            End If
            numStackedConsumables = _maestro.GetStackedConsumablesAtLocation("B1")
            If Not _heightOfSunkenDeck(3) = numStackedConsumables Then
                _rr.errors(3) = True
            End If
        Else
            numStackedConsumables = _maestro.GetStackedConsumablesAtLocation("C0")
            If Not _heightOfSunkenDeck(0) = numStackedConsumables Then
                _rr.errors(0) = True
            End If
            numStackedConsumables = _maestro.GetStackedConsumablesAtLocation("D0")
            If Not _heightOfSunkenDeck(2) = numStackedConsumables Then
                _rr.errors(1) = True
            End If
            numStackedConsumables = _maestro.GetStackedConsumablesAtLocation("C1")
            If Not _heightOfSunkenDeck(3) = numStackedConsumables Then
                _rr.errors(2) = True
            End If
            numStackedConsumables = _maestro.GetStackedConsumablesAtLocation("D1")
            If Not _heightOfSunkenDeck(3) = numStackedConsumables Then
                _rr.errors(3) = True
            End If
        End If
    End Sub

    '' NOTE: No longer always displaying these GUIs (i.e. not on success)...
    '' There are new version of these subroutines that get called on error
    ''
    ''Public Sub DiagnoseHeight()
    ''    Dim hDia As HeightDiagnose = _ph.getDiagnose()
    ''    Dim form As New HeightForm
    ''    form.ShowHeight(hDia, _heightOf384)
    ''    Application.Run(form)
    ''End Sub
    ''Public Sub DiagnoseRec(ByVal loc As String())
    ''    Dim rDia As RecDiagnose = _rf.getDiagnose()
    ''    Dim form As New Form1
    ''    form.ShowRec(rDia, _rr, loc)
    ''    Application.Run(form)
    ''End Sub
    ''Public Sub DiagnoseTip()
    ''    Dim tDia As TipDiagnose = _td.GetDiagnose()
    ''    Dim form As New TipsForm
    ''    form.ShowTip(tDia)
    ''    Application.Run(form)
    ''End Sub
    ''Public Sub DiagnoseNotch_96()
    ''    Dim nDia As NotchDiagnose = _nd.getDiagnose()
    ''    Dim form As New Notch96Form
    ''    form.ShowNotch(nDia, _notchResult)
    ''    Application.Run(form)
    ''End Sub
    ''Public Sub DiagnoseNotch()
    ''    Dim nDia As NotchDiagnose = _nd.getDiagnose()
    ''    Dim form As New NotchForm
    ''    form.ShowNotch(nDia, _notchResult)
    ''    Application.Run(form)
    ''End Sub
    ''Public Sub DiagnoseSunkenDeck()
    ''    Dim sDia As SunkenDeckDiagnose = _sd.GetDiagnose()
    ''    Dim form As New SunkenDeckForm
    ''    form.ShowSunkenDeck(sDia, _heightOfSunkenDeck)
    ''    Application.Run(form)
    ''End Sub

    'release the camera capture instances of all detections
    Public Sub Release()
        _rf.Release()
        _td.Release()
        _ph.Release()
        _nd.Release()
        _sd.Release()
        If Not _capture Is Nothing Then
            _capture.Dispose()
        End If
    End Sub

    'translate the consumable id to a name
    Public Function Consumable(ByVal id As Double) As String
        Dim r As Double = id

        If (id = 0) Then
            Return "Empty"
        ElseIf (id = 0.1) Then
            Return "Unexpected Consumable"
        ElseIf (id = 0.2) Then
            Return "Missing Lid"
        ElseIf (id = 0.3) Then
            Return "Unexpected Lid"
        ElseIf (id = 0.4) Then
            Return "As Expected, No Consumable"
        ElseIf (id >= 1 And id <= 1.4) Then
            If (id = 1.2) Then
                Return "2 * Plate 96"
            ElseIf (id = 1.3) Then
                Return "3 * Plate 96"
            ElseIf (id = 1.4) Then
                Return "4 * Plate 96"
            Else
                Return "Plate 96"
            End If
        ElseIf (id > 1.9 And id < 3) Then
            Return "Tip Box 96"
        ElseIf (id > 2.9 And id < 4) Then
            Return "Plate 384"
        ElseIf (id = 4.1) Then
            Return "Reservoir 1"
        ElseIf (id = 4.4) Then
            Return "Reservoir 4"
        ElseIf (id = 4.12) Then
            Return "Reservoir 12"
        ElseIf (id = 5 Or id = 5.1) Then
            Return "Deep Well Plate 96"
        Else
            Return "Unknown"
        End If
    End Function

    Public Function VisionDoLightingEvaluation(ByRef val As String, ByRef quadVals As String) As String
        _lightingEval.ProcessLightingEvaluation(99, _capture, val, quadVals)
    End Function

    Public Sub SetupCameraProperties(ByVal brightness As Integer)
        StaticClass.SetupCameraProperties(brightness, 4, 0, 99, 2)
    End Sub

#Region "Store Vision Data"

    Public Sub storeDiagnoseSunkenDeck(ByVal deckLocation As String, ByRef visionFeatureResults As Collection, ByVal logToDB As Boolean)
        Dim visionError As Boolean = False ' vision error or not
        Dim connDB As CVisionConnectDB = Nothing
        Dim errorStr As String = String.Empty
        Try
            Dim sDia As SunkenDeckDiagnose = _sd.GetDiagnose()
            Dim vfd As CVisionFeature

            connDB = New CVisionConnectDB

            ' using rr here because the problem is similar to recognition
            If (_rr.errors(0) = True Or _rr.errors(1) = True Or _rr.errors(2) = True Or _rr.errors(3) = True) Then
                visionError = True
            End If

            Dim hsd As String = _heightOfSunkenDeck(0).ToString() & "|" & _
                _heightOfSunkenDeck(1).ToString() & "|" & _
                _heightOfSunkenDeck(2).ToString() & "|" & _
                _heightOfSunkenDeck(3).ToString()
            vfd = New CVisionFeature(StaticClass.FEATURE_DESIGNATION_SUNKENDECK, deckLocation, hsd, Nothing, visionError, _
                CType(sDia, SunkenDeckDiagnose), _
                CType(sDia.userImage(0), Object), _
                CType(sDia.userImage(1), Object), _
                CType(sDia.userImage(2), Object), _
                CType(sDia.userImage(3), Object))

            visionFeatureResults.Add(vfd)

            If logToDB = True Then
                connDB.UpdateVisionDataTable(Guid.NewGuid().ToString(), vfd, errorStr)
            End If
        Finally
            If (Not connDB Is Nothing) Then
                connDB.CloseSQLDBConn(errorStr)
            End If
        End Try
    End Sub

    Public Sub storeRec(ByVal deckLocation As String, ByVal locValues As String, ByRef visionFeatureResults As Collection, ByVal logToDB As Boolean)
        Dim visionError As Boolean = False ' vision error or not
        Dim connDB As CVisionConnectDB = Nothing
        Dim errorStr As String = String.Empty
        Try
            Dim rDia As RecDiagnose = _rf.getDiagnose()
            Dim vfd As CVisionFeature

            connDB = New CVisionConnectDB()

            If (_rr.errors(0) = True Or _rr.errors(1) = True Or _rr.errors(2) = True Or _rr.errors(3) = True) Then
                visionError = True
            End If

            ' squirrel away the diagnostic data in an object that is saved for user
            ' interaction and put in the database for 21 CFR Part 11 satisfaction
            vfd = New CVisionFeature(StaticClass.FEATURE_DESIGNATION_RECOGNITION, deckLocation, locValues, _rr.Copy(), visionError, _
                CType(rDia, RecDiagnose), _
                CType(rDia.userImage(0), Object), _
                CType(rDia.userImage(1), Object), _
                CType(rDia.userImage(2), Object), _
                CType(rDia.userImage(3), Object))

            visionFeatureResults.Add(vfd)

            If logToDB = True Then
                connDB.UpdateVisionDataTable(Guid.NewGuid().ToString(), vfd, errorStr)
            End If
        Finally
            If (Not connDB Is Nothing) Then
                connDB.CloseSQLDBConn(errorStr)
            End If
        End Try
    End Sub

    Public Sub StoreNotch(ByVal deckLocation As String, ByVal expectedNotch As String, ByRef visionFeatureResults As Collection, ByVal logToDB As Boolean)
        Dim visionError As Boolean = False ' vision error or not
        Dim connDB As CVisionConnectDB = Nothing
        Dim errorStr As String = String.Empty
        Try
            Dim nDia As NotchDiagnose = _nd.getDiagnose()
            Dim vfd As CVisionFeature

            connDB = New CVisionConnectDB()

            If _notchResult.Contains(WRONG_NOTCH_ORIENTATION) Then
                visionError = True
            End If

            ' squirrel away the diagnostic data in an object that is saved for user
            ' interaction and put in the database for 21 CFR Part 11 satisfaction
            Dim copyOfUserImage1 As Object = Nothing
            Dim copyOfUserImage2 As Object = Nothing
            Dim copyOfUserImage3 As Object = Nothing
            Dim copyOfUserImage4 As Object = Nothing
            If Not nDia.userImage(1) Is Nothing Then
                copyOfUserImage1 = nDia.userImage(3).Copy()
                copyOfUserImage2 = nDia.userImage(2).Copy()
                copyOfUserImage3 = nDia.userImage(1).Copy()
                copyOfUserImage4 = nDia.userImage(0).Copy()
            Else
                copyOfUserImage1 = nDia.userImage(0).Copy()
            End If
            vfd = New CVisionFeature(StaticClass.FEATURE_DESIGNATION_NOTCH, deckLocation, _notchResult, Nothing, visionError, _
                CType(nDia, NotchDiagnose), _
                copyOfUserImage1, _
                copyOfUserImage2, _
                copyOfUserImage3, _
                copyOfUserImage4)

            visionFeatureResults.Add(vfd)

            If logToDB = True Then
                connDB.UpdateVisionDataTable(Guid.NewGuid().ToString(), vfd, errorStr)
            End If
        Finally
            If (Not connDB Is Nothing) Then
                connDB.CloseSQLDBConn(errorStr)
            End If
        End Try
    End Sub

    Public Sub storeTips(ByVal deckLocation As String, ByVal expectedValues As String, ByRef visionFeatureResults As Collection, ByVal logToDB As Boolean)
        Dim visionError As Boolean = False ' vision error or not
        Dim connDB As CVisionConnectDB = Nothing
        Dim errorStr As String = String.Empty
        Try
            Dim tDia As TipDiagnose = _td.GetDiagnose
            Dim vfd As CVisionFeature

            connDB = New CVisionConnectDB()

            ' squirrel away the diagnostic data in an object that is saved for user
            ' interaction and put in the database for 21 CFR Part 11 satisfaction
            Dim copyOfRBGImage As Object = tDia.rgbImage.Copy()
            vfd = New CVisionFeature(StaticClass.FEATURE_DESIGNATION_TIP, deckLocation, expectedValues, Nothing, _visionErrorInTipDetection, _
                CType(tDia, TipDiagnose), _
                copyOfRBGImage, _
                Nothing, _
                Nothing, _
                Nothing)

            visionFeatureResults.Add(vfd)

            If logToDB = True Then
                connDB.UpdateVisionDataTable(Guid.NewGuid().ToString(), vfd, errorStr)
            End If
        Finally
            If (Not connDB Is Nothing) Then
                connDB.CloseSQLDBConn(errorStr)
            End If
        End Try
    End Sub

    Public Sub storeHeight(ByVal deckLocation As String, ByVal expectedValue As String, ByRef visionFeatureResults As Collection, ByVal logToDB As Boolean)
        Dim visionError As Boolean = False ' vision error or not
        Dim connDB As CVisionConnectDB = Nothing
        Dim errorStr As String = String.Empty
        Try
            Dim hDia As HeightDiagnose = _ph.getDiagnose
            Dim vfd As CVisionFeature

            connDB = New CVisionConnectDB()

            If _rr.errors(0) = True Then
                visionError = True
            End If

            ' squirrel away the diagnostic data in an object that is saved for user
            ' interaction and put in the database for 21 CFR Part 11 satisfaction
            Dim copyOfUserImage As Object = hDia.userImage.Copy()
            vfd = New CVisionFeature(StaticClass.FEATURE_DESIGNATION_HEIGHT384, deckLocation, _
                (_maestro.GetStackedConsumablesAtLocation(deckLocation).ToString() & "," & _heightOf384.ToString()), Nothing, visionError, _
                CType(hDia, HeightDiagnose), _
                copyOfUserImage, _
                Nothing, _
                Nothing, _
                Nothing)

            visionFeatureResults.Add(vfd)

            If logToDB = True Then
                connDB.UpdateVisionDataTable(Guid.NewGuid().ToString(), vfd, errorStr)
            End If
        Finally
            If (Not connDB Is Nothing) Then
                connDB.CloseSQLDBConn(errorStr)
            End If
        End Try
    End Sub

#End Region

#Region "Display Vision (issue) Results"

    '
    ' Now, interact with the user showing diagnostic data
    '
    Public Sub DiagnoseHeight(ByRef vfd As CVisionFeature)
        Dim form As HeightForm
        Try
            Dim hDia As HeightDiagnose = CType(vfd.Diagnose, HeightDiagnose)
            form = New HeightForm
            form.Text = "Height, Deck Location: " & vfd.DeckLocation
            'form.ShowHeight(hDia, _heightOf384)
            form.ShowHeightVF(vfd)
            form.ShowDialog(_maestro.HostWindow)
        Catch ex As Exception
            MsgBox("Error in DiagnoseHeight()" & ex.Message)
        Finally
            form = Nothing
        End Try
    End Sub

    Public Sub DiagnoseRec(ByRef vfd As CVisionFeature)
        Dim form As Form1
        Try
            Dim locArr() As String = vfd.ExpectedOrActualValues.Split("|") ' expected
            If locArr.Length = 4 Then
                Dim rDia As RecDiagnose = CType(vfd.Diagnose, RecDiagnose)
                form = New Form1
                form.Text = "Recognition, Deck Location: " & vfd.DeckLocation
                form.ShowRec(rDia, vfd.RR, locArr)
                form.ShowDialog(_maestro.HostWindow)
            Else
                MsgBox("Unable to DiagnoseRec for deck location: " & vfd.DeckLocation)
            End If
        Catch ex As Exception
            MsgBox("Error in DiagnoseRec()" & ex.Message)
        Finally
            form = Nothing
        End Try
    End Sub

    Public Sub DiagnoseTip(ByRef vfd As CVisionFeature)
        Dim form As TipsForm
        Try
            Dim tDia As TipDiagnose = CType(vfd.Diagnose, TipDiagnose)
            form = New TipsForm
            form.Text = "Tip Detection, Deck Location: " & vfd.DeckLocation
            'form.ShowTip(tDia)
            form.ShowTipVF(vfd, vfd.ErrorFlag)
            form.ShowDialog(_maestro.HostWindow)
        Catch ex As Exception
            MsgBox("Error in DiagnoseTip()" & ex.Message)
        Finally
            form = Nothing
        End Try
    End Sub

    Public Sub DiagnoseNotch_96(ByRef vfd As CVisionFeature)
        Dim form As Notch96Form
        Try
            Dim nDia As NotchDiagnose = CType(vfd.Diagnose, NotchDiagnose)
            form = New Notch96Form
            form.Text = "Notch Detection, Deck Location: " & vfd.DeckLocation
            'form.ShowNotch(nDia, vfd.ExpectedOrActualValues)
            form.ShowNotchVF(vfd)
            form.ShowDialog(_maestro.HostWindow)
        Catch ex As Exception
            MsgBox("Error in DiagnoseNotch_96()" & ex.Message)
        Finally
            form = Nothing
        End Try
    End Sub

    Public Sub DiagnoseNotch(ByRef vfd As CVisionFeature)
        Dim form As NotchForm
        Try
            Dim nDia As NotchDiagnose = CType(vfd.Diagnose, NotchDiagnose)
            form = New NotchForm
            form.Text = "Notch Detection, Deck Location: " & vfd.DeckLocation
            'form.ShowNotch(nDia, vfd.ExpectedOrActualValues)
            form.ShowNotchVF(vfd)
            form.ShowDialog(_maestro.HostWindow)
        Catch ex As Exception
            MsgBox("Error in DiagnoseNotch()" & ex.Message)
        Finally
            form = Nothing
        End Try
    End Sub

    Public Sub DiagnoseSunkenDeck(ByRef vfd As CVisionFeature)
        Dim form As SunkenDeckForm
        Try
            Dim heights() As Integer = New Integer() {0, 0, 0, 0}
            Dim strArr() As String = vfd.ExpectedOrActualValues.Split("|") ' expected
            If strArr.Length = 4 Then
                heights(0) = Convert.ToInt32(strArr(0))
                heights(1) = Convert.ToInt32(strArr(1))
                heights(2) = Convert.ToInt32(strArr(2))
                heights(3) = Convert.ToInt32(strArr(3))
            End If

            Dim sDia As SunkenDeckDiagnose = CType(vfd.Diagnose, SunkenDeckDiagnose)
            form = New SunkenDeckForm
            form.Text = "Sunken Deck, Deck Location: " & vfd.DeckLocation
            form.ShowSunkenDeck(sDia, heights)
            form.ShowDialog(_maestro.HostWindow)
        Catch ex As Exception
            MsgBox("Error in DiagnoseSunkenDeck()" & ex.Message)
        Finally
            form = Nothing
        End Try
    End Sub

#End Region

End Class